import os
from ..utils.util import randstr, APP_ID
from ..utils.log import *
import aiohttp
import asyncio
import requests
from typing import Dict, Any, TypedDict, Optional, Literal
from urllib.parse import quote
import time

_APP_NAME = os.getenv("APP_NAME", "ieops")
_SERVER_SOCKET_DIR = os.getenv("SERVER_SOCKET_DIR", os.getcwd()) # 默认是项目根目录

_MODEL_THREAD_CONCURRENCY = os.getenv("MODEL_THREAD_CONCURRENCY", "8")
_REGISTER_INTERVAL = os.getenv("REGISTER_INTERVAL", "10")
_REGISTER_ENABLED = os.getenv("REGISTER_ENABLED", "true")

# connector type
ConnectorType = Literal["proxy", "brslet"]

# socket 配置
_SOCKET_PATHS: Dict[ConnectorType, str] = {
    "proxy": os.getenv("IEOPS_PROXY_SOCKET", "/var/run/ieops/proxy.sock"),
    "brslet": os.getenv("IEOPS_BRSLET_SOCKET", "/var/run/ieops/brslet.sock"),
}

# async unix connectors (lazy init)
_CONNECTORS: Dict[ConnectorType, Optional[aiohttp.UnixConnector]] = {
    "proxy": None,
    "brslet": None,
}

# sync session (lazy init)
_SYNC_SESSION: Optional[requests.Session] = None


def _build_unix_url(socket_path: str, api_path: str) -> str:
    """构建 requests_unixsocket 格式的 Unix socket URL
    
    Args:
        socket_path: socket 文件路径，如 /var/run/ieops/proxy.sock
        api_path: API 路径，如 register
        
    Returns:
        http+unix://%2Fvar%2Frun%2Fieops%2Fproxy.sock/register
    """
    encoded_socket = quote(socket_path, safe='')
    return f"http+unix://{encoded_socket}/{api_path}"


class Payload(TypedDict):
    path: str
    payload: Dict[str, Any]

class Response:
    """HTTP 响应封装"""
    def __init__(self, message: str, data: Optional[Any] = None):
        self.message = message
        self.data = data
    
    def __repr__(self) -> str:
        return f"Response(message={self.message!r}, data={self.data!r})"


def _get_connector(connector_type: ConnectorType) -> aiohttp.UnixConnector:
    """获取指定类型的异步 Unix socket connector（懒加载）"""
    if _CONNECTORS[connector_type] is None:
        _CONNECTORS[connector_type] = aiohttp.UnixConnector(path=_SOCKET_PATHS[connector_type])
    return _CONNECTORS[connector_type]


def _get_sync_session() -> requests.Session:
    """获取同步 HTTP session，支持 Unix socket"""
    global _SYNC_SESSION
    if _SYNC_SESSION is None:
        _SYNC_SESSION = requests.Session()
        # 添加 Unix socket 支持
        try:
            from requests_unixsocket import UnixAdapter
            _SYNC_SESSION.mount("http+unix://", UnixAdapter())
        except ImportError:
            uvicorn_logger.warning("requests_unixsocket not installed, Unix socket support disabled")
    return _SYNC_SESSION


def post(_payload: Payload, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """send POST request via Unix socket
    
    Args:
        _payload: 请求数据
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    session = _get_sync_session()
    url = _build_unix_url(_SOCKET_PATHS[connector_type], _payload['path'])
    for attempt in range(max_retries + 1):
        try:
            resp = session.post(url, json=_payload['payload'], timeout=timeout)
            if resp.status_code == 200:
                try:
                    data = resp.json()
                    return Response(message=data.get('message', ''), data=data.get('data'))
                except ValueError as e:
                    err = Response(message=f"Invalid JSON response: {e}")
            else:
                err = Response(message=f"HTTP {resp.status_code}: {resp.text[:200]}")
        except requests.Timeout as e:
            err = Response(message=f"Client POST [{_payload['path']}] timeout: {e}")
        except requests.RequestException as e:
            err = Response(message=f"Client POST [{_payload['path']}] failed: {e}")
        if attempt < max_retries:
            time.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client POST [{_payload['path']}] failed after {max_retries + 1} attempts: {err}")
    return err


def get(path: str, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """send GET request via Unix socket
    
    Args:
        path: API 路径
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    session = _get_sync_session()
    url = _build_unix_url(_SOCKET_PATHS[connector_type], path)
    for attempt in range(max_retries + 1):
        try:
            resp = session.get(url, timeout=timeout)
            if resp.status_code == 200:
                try:
                    data = resp.json()
                    return Response(message=data.get('message', ''), data=data.get('data'))
                except ValueError as e:
                    err = Response(message=f"Invalid JSON response: {e}")
            else:
                err = Response(message=f"HTTP {resp.status_code}: {resp.text[:200]}")
        except requests.Timeout as e:
            err = Response(message=f"Client GET [{path}] timeout: {e}")
        except requests.RequestException as e:
            err = Response(message=f"Client GET [{path}] failed: {e}")
        if attempt < max_retries:
            time.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client GET [{path}] failed after {max_retries + 1} attempts: {err}")
    return err


def delete(path: str, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """send DELETE request via Unix socket
    
    Args:
        path: API 路径
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    session = _get_sync_session()
    url = _build_unix_url(_SOCKET_PATHS[connector_type], path)
    for attempt in range(max_retries + 1):
        try:
            resp = session.delete(url, timeout=timeout)
            if resp.status_code == 200:
                try:
                    data = resp.json()
                    return Response(message=data.get('message', ''), data=data.get('data'))
                except ValueError as e:
                    err = Response(message=f"Invalid JSON response: {e}")
            else:
                err = Response(message=f"HTTP {resp.status_code}: {resp.text[:200]}")
        except requests.Timeout as e:
            err = Response(message=f"Client DELETE [{path}] timeout: {e}")
        except requests.RequestException as e:
            err = Response(message=f"Client DELETE [{path}] failed: {e}")
        if attempt < max_retries:
            time.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client DELETE [{path}] failed after {max_retries + 1} attempts: {err}")
    return err


async def async_post(_payload: Payload, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """async POST request via Unix socket
    
    Args:
        _payload: 请求数据
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    url = f"http://localhost/{_payload['path']}"  # aiohttp UnixConnector 使用 localhost
    client_timeout = aiohttp.ClientTimeout(total=timeout)
    
    for attempt in range(max_retries + 1):
        try:
            async with aiohttp.ClientSession(
                connector=_get_connector(connector_type), 
                connector_owner=False,
                timeout=client_timeout
            ) as session:
                async with session.post(url, json=_payload['payload']) as resp:
                    if resp.status == 200:
                        try:
                            data = await resp.json()
                            return Response(message=data.get('message', ''), data=data.get('data'))
                        except (ValueError, aiohttp.ContentTypeError) as e:
                            err = Response(message=f"Invalid JSON response: {e}")
                    else:
                        text = await resp.text()
                        err = Response(message=f"HTTP {resp.status}: {text[:200]}")
        except asyncio.TimeoutError:
            err = Response(message=f"Client POST [{_payload['path']}] timeout")
        except aiohttp.ClientError as e:
            err = Response(message=f"Client POST [{_payload['path']}] failed: {e}")
        except asyncio.CancelledError:
            raise  # 重新抛出，让调用者处理取消
        if attempt < max_retries:
            await asyncio.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client POST [{_payload['path']}] failed after {max_retries + 1} attempts: {err}")
    return err


async def async_get(path: str, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """async GET request via Unix socket
    
    Args:
        path: API 路径
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    url = f"http://localhost/{path}"  # aiohttp UnixConnector 使用 localhost
    client_timeout = aiohttp.ClientTimeout(total=timeout)
    
    for attempt in range(max_retries + 1):
        try:
            async with aiohttp.ClientSession(
                connector=_get_connector(connector_type), 
                connector_owner=False,
                timeout=client_timeout
            ) as session:
                async with session.get(url) as resp:
                    if resp.status == 200:
                        try:
                            data = await resp.json()
                            return Response(message=data.get('message', ''), data=data.get('data'))
                        except (ValueError, aiohttp.ContentTypeError) as e:
                            err = Response(message=f"Invalid JSON response: {e}")
                    else:
                        text = await resp.text()
                        err = Response(message=f"HTTP {resp.status}: {text[:200]}")
        except asyncio.TimeoutError:
            err = Response(message=f"Client GET [{path}] timeout")
        except aiohttp.ClientError as e:
            err = Response(message=f"Client GET [{path}] failed: {e}")
        except asyncio.CancelledError:
            raise  # 重新抛出，让调用者处理取消
        if attempt < max_retries:
            await asyncio.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client GET [{path}] failed after {max_retries + 1} attempts: {err}")
    return err


async def async_delete(path: str, timeout: float = 0.8, max_retries: int = 2, connector_type: ConnectorType = "proxy") -> Response:
    """async DELETE request via Unix socket
    
    Args:
        path: API 路径
        timeout: 超时时间
        max_retries: 最大重试次数
        connector_type: "proxy" 或 "brslet"
    """
    err = None
    url = f"http://localhost/{path}"  # aiohttp UnixConnector 使用 localhost
    client_timeout = aiohttp.ClientTimeout(total=timeout)
    
    for attempt in range(max_retries + 1):
        try:
            async with aiohttp.ClientSession(
                connector=_get_connector(connector_type), 
                connector_owner=False,
                timeout=client_timeout
            ) as session:
                async with session.delete(url) as resp:
                    if resp.status == 200:
                        try:
                            data = await resp.json()
                            return Response(message=data.get('message', ''), data=data.get('data'))
                        except (ValueError, aiohttp.ContentTypeError) as e:
                            err = Response(message=f"Invalid JSON response: {e}")
                    else:
                        text = await resp.text()
                        err = Response(message=f"HTTP {resp.status}: {text[:200]}")
        except asyncio.TimeoutError:
            err = Response(message=f"Client DELETE [{path}] timeout")
        except aiohttp.ClientError as e:
            err = Response(message=f"Client DELETE [{path}] failed: {e}")
        except asyncio.CancelledError:
            raise  # 重新抛出，让调用者处理取消
        if attempt < max_retries:
            await asyncio.sleep(0.1 * (attempt + 1))
        else:
            uvicorn_logger.warning(f"Client DELETE [{path}] failed after {max_retries + 1} attempts: {err}")
    return err

class Register:
    def __init__(self):
        self._appinfo = self._app_info()
        
    def _app_info(self):
        self._tokens = [randstr(8) for _ in range(int(_MODEL_THREAD_CONCURRENCY))]
        return {
            "id": APP_ID,
            "server_socket": f"{_SERVER_SOCKET_DIR}/{APP_ID}.sock",
            "max_concurrent_reqs": int(_MODEL_THREAD_CONCURRENCY),
            "endpoint": _APP_NAME,
            "weight": 1,
        }

    async def _unregister_old(self):
        """启动时先取消旧的注册，清理可能存在的僵尸注册
        按 socket 目录清理，避免误伤同 endpoint 但不同 pod 的服务
        """
        socket_dir = _SERVER_SOCKET_DIR
        uvicorn_logger.info("Unregistering old models in socket_dir {} from IEOPS proxy (if exists)...".format(socket_dir))
        # 取消同一目录下的旧注册（同一 pod 重启后目录相同，但 socket ID 不同）
        await async_post(Payload(path="unregister", payload={"socket_dir": socket_dir}), max_retries=0)

    async def _unregister(self):
        """停止时取消当前注册"""
        uvicorn_logger.info("Unregistering {} from IEOPS proxy...".format(_APP_NAME))
        await async_post(Payload(path="unregister", payload={"id": APP_ID}), max_retries=1)

    async def register(self):
        if _REGISTER_ENABLED.lower() == "false":
            uvicorn_logger.info("Register is disabled, skipping...")
            return
        uvicorn_logger.info("Registering {} to IEOPS proxy...".format(_APP_NAME))
        try:
            # 方案3: 启动时先清理旧注册
            await self._unregister_old()
            while True:
                await async_post(Payload(path="register", payload=self._appinfo))
                await asyncio.sleep(int(_REGISTER_INTERVAL))
        except asyncio.CancelledError:
            uvicorn_logger.info("Register stopped")
            raise
        finally:
            # 停止时取消注册
            try:
                await self._unregister()
            except Exception as e:
                uvicorn_logger.warning(f"Failed to unregister: {e}")
            # 关闭所有 connectors
            for connector in _CONNECTORS.values():
                if connector is not None:
                    await connector.close()
            if _SYNC_SESSION is not None:
                _SYNC_SESSION.close()

__all__ = [
    'Payload', 'Response', 'ConnectorType',
    'post', 'get', 'delete', 
    'async_post', 'async_get', 'async_delete', 
    'Register',
]


# /root/code/python/ieops-python/sdk/src/handler/server/diffuser.py
"""
Diffuser Server - Hugging Face Diffusers 推理服务

直接集成 Diffusers 库，无需子进程。提供：
- 文生图 (txt2img)
- 图生图 (img2img)
- 图像修复 (inpaint)
- 流式进度回调
- LoRA 加载

架构（直接集成，无子进程）：
┌─────────────────┐     ┌────────────────────────────────┐
│  外部客户端      │────▶│  DiffuserServer (FastAPI)      │
│  (HTTP/SSE)     │     │  内置 Diffusers Pipeline       │
└─────────────────┘     └────────────────────────────────┘
"""

import asyncio
import json
import os
import base64
import gc
import time
import uuid
from io import BytesIO
from typing import Any, AsyncGenerator, Dict, List, Optional, Callable
from contextlib import asynccontextmanager

import uvicorn
from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import StreamingResponse, JSONResponse
from pydantic import BaseModel, Field

from .server import BaseServer
from ...utils.log import uvicorn_logger, log
from ...utils.util import APP_ID, arandstr

# ========== 环境变量配置 ==========
_MODEL_SERVER_SOCKET = os.getenv("MODEL_SERVER_SOCKET", f"{APP_ID}.sock")
_MODEL_SERVER_PORT = int(os.getenv("MODEL_SERVER_PORT", 8001))
_MODEL_SERVER_HOST = os.getenv("MODEL_SERVER_HOST", "127.0.0.1")
_GRACEFUL_SHUTDOWN_TIME = int(os.getenv("GRACEFUL_SHUTDOWN_TIME", 3))
_MODEL_TIMEOUT = int(os.getenv("MODEL_TIMEOUT", 600))

# Diffuser 配置
_MODEL_PATH = os.getenv("MODEL_PATH", "stabilityai/stable-diffusion-xl-base-1.0")
_MODEL_TYPE = os.getenv("MODEL_TYPE", "auto")
_TORCH_DTYPE = os.getenv("TORCH_DTYPE", "float16")
_DEVICE = os.getenv("DEVICE", "auto")
_ENABLE_VAE_SLICING = os.getenv("ENABLE_VAE_SLICING", "true").lower() == "true"
_ENABLE_VAE_TILING = os.getenv("ENABLE_VAE_TILING", "false").lower() == "true"
_ENABLE_MODEL_CPU_OFFLOAD = os.getenv("ENABLE_MODEL_CPU_OFFLOAD", "false").lower() == "true"
_ENABLE_XFORMERS = os.getenv("ENABLE_XFORMERS", "false").lower() == "true"
_COMPILE_MODEL = os.getenv("COMPILE_MODEL", "false").lower() == "true"

SERVER_CODE_OK = 0
SERVER_CODE_ERROR = 1
SERVER_CODE_STOP = 2


# ========== 延迟导入 Diffusers（避免未安装时报错）==========
def _import_diffusers():
    """延迟导入 diffusers 相关模块"""
    import torch
    from PIL import Image
    from diffusers import (
        AutoPipelineForText2Image,
        AutoPipelineForImage2Image,
        AutoPipelineForInpainting,
        DPMSolverMultistepScheduler,
        EulerAncestralDiscreteScheduler,
        EulerDiscreteScheduler,
        DDIMScheduler,
        UniPCMultistepScheduler,
        LCMScheduler,
    )
    return {
        "torch": torch,
        "Image": Image,
        "AutoPipelineForText2Image": AutoPipelineForText2Image,
        "AutoPipelineForImage2Image": AutoPipelineForImage2Image,
        "AutoPipelineForInpainting": AutoPipelineForInpainting,
        "schedulers": {
            "dpm++_2m": DPMSolverMultistepScheduler,
            "dpm++_2m_karras": DPMSolverMultistepScheduler,
            "euler": EulerDiscreteScheduler,
            "euler_a": EulerAncestralDiscreteScheduler,
            "ddim": DDIMScheduler,
            "unipc": UniPCMultistepScheduler,
            "lcm": LCMScheduler,
        }
    }


# ========== Pydantic 模型定义 ==========
class Txt2ImgRequest(BaseModel):
    """文生图请求"""
    prompt: str = Field(..., description="正向提示词")
    negative_prompt: str = Field(default="", description="负向提示词")
    width: int = Field(default=1024, ge=256, le=4096, description="图像宽度")
    height: int = Field(default=1024, ge=256, le=4096, description="图像高度")
    num_inference_steps: int = Field(default=30, ge=1, le=150, description="推理步数")
    guidance_scale: float = Field(default=7.5, ge=0, le=30, description="引导系数")
    num_images: int = Field(default=1, ge=1, le=4, description="生成图片数量")
    seed: Optional[int] = Field(default=None, description="随机种子")
    scheduler: Optional[str] = Field(default=None, description="调度器类型")
    lora_weights: Optional[List[Dict[str, Any]]] = Field(default=None, description="LoRA 权重配置")


class Img2ImgRequest(BaseModel):
    """图生图请求"""
    prompt: str = Field(..., description="正向提示词")
    negative_prompt: str = Field(default="", description="负向提示词")
    image: str = Field(..., description="输入图像 (base64)")
    strength: float = Field(default=0.75, ge=0, le=1, description="去噪强度")
    num_inference_steps: int = Field(default=30, ge=1, le=150, description="推理步数")
    guidance_scale: float = Field(default=7.5, ge=0, le=30, description="引导系数")
    num_images: int = Field(default=1, ge=1, le=4, description="生成图片数量")
    seed: Optional[int] = Field(default=None, description="随机种子")


class InpaintRequest(BaseModel):
    """图像修复请求"""
    prompt: str = Field(..., description="正向提示词")
    negative_prompt: str = Field(default="", description="负向提示词")
    image: str = Field(..., description="输入图像 (base64)")
    mask_image: str = Field(..., description="遮罩图像 (base64)")
    width: Optional[int] = Field(default=None, description="输出宽度")
    height: Optional[int] = Field(default=None, description="输出高度")
    num_inference_steps: int = Field(default=30, ge=1, le=150, description="推理步数")
    guidance_scale: float = Field(default=7.5, ge=0, le=30, description="引导系数")
    strength: float = Field(default=0.75, ge=0, le=1, description="去噪强度")
    num_images: int = Field(default=1, ge=1, le=4, description="生成图片数量")
    seed: Optional[int] = Field(default=None, description="随机种子")


class DiffuserServer(BaseServer):
    """
    Diffuser 推理服务器 - 直接集成 Diffusers 库
    
    无需子进程，直接在当前进程中加载和运行 Diffusers Pipeline。
    """
    
    def __init__(
        self,
        model_path: Optional[str] = None,
        model_type: Optional[str] = None,
        torch_dtype: Optional[str] = None,
        device: Optional[str] = None,
        on_request_received: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        enable_vae_slicing: Optional[bool] = None,
        enable_vae_tiling: Optional[bool] = None,
        enable_model_cpu_offload: Optional[bool] = None,
        enable_xformers: Optional[bool] = None,
        compile_model: Optional[bool] = None,
    ) -> None:
        """
        Args:
            model_path: 模型路径或 HuggingFace 模型 ID
            model_type: 模型类型 (auto, sd15, sdxl, sd3, flux)
            torch_dtype: 数据类型 (float16, float32, bfloat16)
            device: 设备 (cuda, cpu, auto)
            on_request_received: 请求预处理回调
            enable_vae_slicing: 启用 VAE slicing（减少显存）
            enable_vae_tiling: 启用 VAE tiling（支持大图）
            enable_model_cpu_offload: 启用模型 CPU 卸载
            enable_xformers: 启用 xformers 加速
            compile_model: 启用 PyTorch 2.0 compile
        """
        self._model_path = model_path or _MODEL_PATH
        self._model_type = model_type or _MODEL_TYPE
        self._torch_dtype = torch_dtype or _TORCH_DTYPE
        self._device = device or _DEVICE
        self._on_request_received = on_request_received
        
        # 优化选项
        self._enable_vae_slicing = enable_vae_slicing if enable_vae_slicing is not None else _ENABLE_VAE_SLICING
        self._enable_vae_tiling = enable_vae_tiling if enable_vae_tiling is not None else _ENABLE_VAE_TILING
        self._enable_model_cpu_offload = enable_model_cpu_offload if enable_model_cpu_offload is not None else _ENABLE_MODEL_CPU_OFFLOAD
        self._enable_xformers = enable_xformers if enable_xformers is not None else _ENABLE_XFORMERS
        self._compile_model = compile_model if compile_model is not None else _COMPILE_MODEL
        
        # Pipeline（启动时加载）
        self._modules: Optional[Dict[str, Any]] = None
        self._txt2img_pipe = None
        self._img2img_pipe = None
        self._inpaint_pipe = None
        self._loaded_loras: List[str] = []
        self._tasks: Dict[str, Dict[str, Any]] = {}
        
        self._app = FastAPI(lifespan=self._lifespan)
        self._server: Optional[uvicorn.Server] = None
        
        self._setup_routes()

    @asynccontextmanager
    async def _lifespan(self, app: FastAPI):
        """FastAPI 生命周期管理"""
        uvicorn_logger.info("Diffuser Server starting...")
        
        # 加载模型
        await self._load_model()
        
        yield
        
        # 清理
        uvicorn_logger.info("Diffuser Server shutting down...")
        await self._unload_model()
        
        # 清理 socket 文件
        if _MODEL_SERVER_SOCKET and os.path.exists(_MODEL_SERVER_SOCKET):
            os.remove(_MODEL_SERVER_SOCKET)

    async def _load_model(self):
        """加载 Diffusers 模型"""
        uvicorn_logger.info(f"Loading model: {self._model_path}")
        start_time = time.time()
        
        try:
            # 延迟导入
            self._modules = _import_diffusers()
            torch = self._modules["torch"]
            
            # 确定设备
            if self._device == "auto":
                self._device = "cuda" if torch.cuda.is_available() else "cpu"
            
            # 确定数据类型
            dtype_map = {
                "float16": torch.float16,
                "float32": torch.float32,
                "bfloat16": torch.bfloat16,
            }
            torch_dtype = dtype_map.get(self._torch_dtype, torch.float16)
            
            # 加载 txt2img pipeline
            AutoPipelineForText2Image = self._modules["AutoPipelineForText2Image"]
            self._txt2img_pipe = AutoPipelineForText2Image.from_pretrained(
                self._model_path,
                torch_dtype=torch_dtype,
                use_safetensors=True,
                variant="fp16" if torch_dtype == torch.float16 else None,
            )
            
            # 应用优化
            if not self._enable_model_cpu_offload:
                self._txt2img_pipe = self._txt2img_pipe.to(self._device)
            
            if self._enable_vae_slicing and hasattr(self._txt2img_pipe, 'enable_vae_slicing'):
                self._txt2img_pipe.enable_vae_slicing()
                
            if self._enable_vae_tiling and hasattr(self._txt2img_pipe, 'enable_vae_tiling'):
                self._txt2img_pipe.enable_vae_tiling()
                
            if self._enable_model_cpu_offload and hasattr(self._txt2img_pipe, 'enable_model_cpu_offload'):
                self._txt2img_pipe.enable_model_cpu_offload()
                
            if self._enable_xformers:
                try:
                    self._txt2img_pipe.enable_xformers_memory_efficient_attention()
                except Exception as e:
                    uvicorn_logger.warning(f"Could not enable xformers: {e}")
            
            if self._compile_model and hasattr(torch, "compile"):
                if hasattr(self._txt2img_pipe, 'unet'):
                    self._txt2img_pipe.unet = torch.compile(self._txt2img_pipe.unet, mode="reduce-overhead")
            
            # 创建 img2img 和 inpaint pipeline
            try:
                AutoPipelineForImage2Image = self._modules["AutoPipelineForImage2Image"]
                self._img2img_pipe = AutoPipelineForImage2Image.from_pipe(self._txt2img_pipe)
            except Exception as e:
                uvicorn_logger.warning(f"Could not create img2img pipeline: {e}")
            
            try:
                AutoPipelineForInpainting = self._modules["AutoPipelineForInpainting"]
                self._inpaint_pipe = AutoPipelineForInpainting.from_pipe(self._txt2img_pipe)
            except Exception as e:
                uvicorn_logger.warning(f"Could not create inpaint pipeline: {e}")
            
            load_time = time.time() - start_time
            uvicorn_logger.info(f"Model loaded in {load_time:.2f}s")
            
        except Exception as e:
            uvicorn_logger.error(f"Failed to load model: {e}")
            raise

    async def _unload_model(self):
        """卸载模型，释放显存"""
        self._txt2img_pipe = None
        self._img2img_pipe = None
        self._inpaint_pipe = None
        
        gc.collect()
        
        if self._modules:
            torch = self._modules["torch"]
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

    def shutdown(self) -> None:
        """请求服务器优雅关闭"""
        if self._server:
            self._server.should_exit = True

    async def serve(self) -> None:
        """启动服务器"""
        if _MODEL_SERVER_SOCKET:
            config = uvicorn.Config(
                self._app,
                uds=_MODEL_SERVER_SOCKET,
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=_GRACEFUL_SHUTDOWN_TIME
            )
        else:
            config = uvicorn.Config(
                self._app,
                host=_MODEL_SERVER_HOST,
                port=_MODEL_SERVER_PORT,
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=_GRACEFUL_SHUTDOWN_TIME
            )
        
        self._server = uvicorn.Server(config)
        await self._server.serve()

    def _setup_routes(self) -> None:
        """设置 API 路由"""
        
        @self._app.get("/health")
        async def health():
            """健康检查"""
            if self._txt2img_pipe is None:
                return JSONResponse({"status": "initializing"}, status_code=503)
            return {"status": "ok"}

        @self._app.get("/")
        async def root():
            """根路径"""
            return {
                "service": "Diffuser Server",
                "model": self._model_path,
                "status": "ready" if self._txt2img_pipe else "initializing",
                "endpoints": {
                    "health": "/health",
                    "txt2img": "/v1/txt2img",
                    "img2img": "/v1/img2img",
                    "inpaint": "/v1/inpaint",
                    "model_info": "/v1/model/info",
                }
            }

        @self._app.get("/v1/model/info")
        async def get_model_info():
            """获取模型信息"""
            torch = self._modules["torch"] if self._modules else None
            info = {
                "model_path": self._model_path,
                "model_type": self._model_type,
                "device": self._device,
                "torch_dtype": self._torch_dtype,
                "loaded": self._txt2img_pipe is not None,
                "loaded_loras": self._loaded_loras,
                "features": {
                    "txt2img": self._txt2img_pipe is not None,
                    "img2img": self._img2img_pipe is not None,
                    "inpaint": self._inpaint_pipe is not None,
                },
                "optimizations": {
                    "vae_slicing": self._enable_vae_slicing,
                    "vae_tiling": self._enable_vae_tiling,
                    "model_cpu_offload": self._enable_model_cpu_offload,
                    "xformers": self._enable_xformers,
                    "compile": self._compile_model,
                }
            }
            if torch and torch.cuda.is_available():
                info["cuda_memory_allocated"] = f"{torch.cuda.memory_allocated() / 1024**3:.2f} GB"
                info["cuda_memory_reserved"] = f"{torch.cuda.memory_reserved() / 1024**3:.2f} GB"
            return info

        @self._app.get("/v1/schedulers")
        async def get_schedulers():
            """获取可用调度器"""
            schedulers = list(self._modules["schedulers"].keys()) if self._modules else []
            return {"schedulers": schedulers}

        # ========== 文生图 ==========
        @self._app.post("/v1/txt2img")
        async def txt2img(request: Txt2ImgRequest):
            """文生图（同步）"""
            result = None
            async for data in self._txt2img(request, stream=False):
                if data.get("type") == "result":
                    result = data
                elif data.get("type") == "error":
                    raise HTTPException(status_code=500, detail=data["error"])
            return result

        @self._app.post("/v1/txt2img/stream")
        async def txt2img_stream(request: Txt2ImgRequest):
            """文生图（流式）"""
            return StreamingResponse(
                self._txt2img_sse(request),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        # ========== 图生图 ==========
        @self._app.post("/v1/img2img")
        async def img2img(request: Img2ImgRequest):
            """图生图（同步）"""
            result = None
            async for data in self._img2img(request, stream=False):
                if data.get("type") == "result":
                    result = data
                elif data.get("type") == "error":
                    raise HTTPException(status_code=500, detail=data["error"])
            return result

        @self._app.post("/v1/img2img/stream")
        async def img2img_stream(request: Img2ImgRequest):
            """图生图（流式）"""
            return StreamingResponse(
                self._img2img_sse(request),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        # ========== 图像修复 ==========
        @self._app.post("/v1/inpaint")
        async def inpaint(request: InpaintRequest):
            """图像修复（同步）"""
            result = None
            async for data in self._inpaint(request, stream=False):
                if data.get("type") == "result":
                    result = data
                elif data.get("type") == "error":
                    raise HTTPException(status_code=500, detail=data["error"])
            return result

        @self._app.post("/v1/inpaint/stream")
        async def inpaint_stream(request: InpaintRequest):
            """图像修复（流式）"""
            return StreamingResponse(
                self._inpaint_sse(request),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        # ========== 兼容 ieops 协议 ==========
        @self._app.post("/sse/infer")
        async def infer_sse(request: Request):
            """SSE 推理接口"""
            body = await request.body()
            return StreamingResponse(
                self._ieops_sse_generator(body),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        @self._app.post("/infer")
        async def infer(request: Request):
            """同步推理接口"""
            body = await request.body()
            results: List[Any] = []
            last_code = SERVER_CODE_OK
            
            async for item in self._ieops_generator(body):
                code = item["code"]
                data = item["data"]
                if code == SERVER_CODE_ERROR:
                    last_code = code
                    results.append({"error": data})
                    break
                if code == SERVER_CODE_STOP:
                    break
                try:
                    results.append(json.loads(data) if isinstance(data, str) else data)
                except json.JSONDecodeError:
                    results.append(data)
            
            return {"code": last_code, "results": results}

    # ========== 工具方法 ==========
    def _decode_base64_image(self, base64_str: str):
        """解码 base64 图像"""
        Image = self._modules["Image"]
        if base64_str.startswith("data:"):
            base64_str = base64_str.split(",", 1)[1]
        image_data = base64.b64decode(base64_str)
        return Image.open(BytesIO(image_data)).convert("RGB")

    def _encode_image_to_base64(self, image, format: str = "PNG") -> str:
        """编码图像为 base64"""
        buffer = BytesIO()
        image.save(buffer, format=format)
        return base64.b64encode(buffer.getvalue()).decode("utf-8")

    def _set_scheduler(self, pipe, scheduler_name: Optional[str]):
        """设置调度器"""
        if not scheduler_name or not self._modules:
            return
        
        schedulers = self._modules["schedulers"]
        if scheduler_name not in schedulers:
            return
        
        scheduler_class = schedulers[scheduler_name]
        config = pipe.scheduler.config
        
        if "karras" in scheduler_name:
            pipe.scheduler = scheduler_class.from_config(config, use_karras_sigmas=True)
        else:
            pipe.scheduler = scheduler_class.from_config(config)

    def _load_lora(self, pipe, lora_weights: List[Dict[str, Any]]):
        """加载 LoRA"""
        for lora in lora_weights:
            lora_path = lora.get("path")
            lora_scale = lora.get("scale", 1.0)
            adapter_name = lora.get("name", os.path.basename(lora_path) if lora_path else "default")
            
            if lora_path and lora_path not in self._loaded_loras:
                try:
                    pipe.load_lora_weights(lora_path, adapter_name=adapter_name)
                    pipe.set_adapters([adapter_name], adapter_weights=[lora_scale])
                    self._loaded_loras.append(lora_path)
                    uvicorn_logger.info(f"Loaded LoRA: {lora_path}")
                except Exception as e:
                    uvicorn_logger.error(f"Failed to load LoRA: {e}")

    def _create_progress_callback(self, task_id: str, total_steps: int):
        """创建进度回调"""
        def callback(pipe, step_index, timestep, callback_kwargs):
            progress = (step_index + 1) / total_steps
            if task_id in self._tasks:
                self._tasks[task_id]["progress"] = progress
                self._tasks[task_id]["step"] = step_index + 1
            return callback_kwargs
        return callback

    # ========== 推理方法 ==========
    async def _txt2img(self, request: Txt2ImgRequest, stream: bool = False) -> AsyncGenerator[Dict[str, Any], None]:
        """文生图推理"""
        if not self._txt2img_pipe:
            yield {"type": "error", "error": "Model not loaded"}
            return
        
        torch = self._modules["torch"]
        task_id = str(uuid.uuid4())
        self._tasks[task_id] = {"progress": 0, "step": 0}
        
        try:
            # 种子
            seed = request.seed if request.seed is not None else int(torch.randint(0, 2**32 - 1, (1,)).item())
            generator = torch.Generator(device=self._device).manual_seed(seed)
            
            # 调度器
            if request.scheduler:
                self._set_scheduler(self._txt2img_pipe, request.scheduler)
            
            # LoRA
            if request.lora_weights:
                self._load_lora(self._txt2img_pipe, request.lora_weights)
            
            # 进度回调
            callback = self._create_progress_callback(task_id, request.num_inference_steps) if stream else None
            
            # 推理（在线程池中运行避免阻塞）
            loop = asyncio.get_event_loop()
            
            def run_inference():
                return self._txt2img_pipe(
                    prompt=request.prompt,
                    negative_prompt=request.negative_prompt or None,
                    width=request.width,
                    height=request.height,
                    num_inference_steps=request.num_inference_steps,
                    guidance_scale=request.guidance_scale,
                    num_images_per_prompt=request.num_images,
                    generator=generator,
                    callback_on_step_end=callback,
                )
            
            if stream:
                inference_task = loop.run_in_executor(None, run_inference)
                last_step = 0
                while not inference_task.done():
                    await asyncio.sleep(0.1)
                    current_step = self._tasks[task_id].get("step", 0)
                    if current_step > last_step:
                        yield {
                            "type": "progress",
                            "progress": self._tasks[task_id]["progress"],
                            "step": current_step,
                            "total_steps": request.num_inference_steps,
                        }
                        last_step = current_step
                output = await inference_task
            else:
                output = await loop.run_in_executor(None, run_inference)
            
            images_base64 = [self._encode_image_to_base64(img) for img in output.images]
            
            yield {
                "type": "result",
                "images": images_base64,
                "seed": seed,
                "task_id": task_id,
            }
            
        except Exception as e:
            uvicorn_logger.error(f"txt2img error: {e}")
            yield {"type": "error", "error": str(e)}
        finally:
            del self._tasks[task_id]
            gc.collect()
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

    async def _txt2img_sse(self, request: Txt2ImgRequest) -> AsyncGenerator[str, None]:
        """文生图 SSE 生成器"""
        async for data in self._txt2img(request, stream=True):
            yield f"data: {json.dumps(data)}\n\n"

    async def _img2img(self, request: Img2ImgRequest, stream: bool = False) -> AsyncGenerator[Dict[str, Any], None]:
        """图生图推理"""
        if not self._img2img_pipe:
            yield {"type": "error", "error": "img2img pipeline not available"}
            return
        
        torch = self._modules["torch"]
        task_id = str(uuid.uuid4())
        self._tasks[task_id] = {"progress": 0, "step": 0}
        
        try:
            init_image = self._decode_base64_image(request.image)
            seed = request.seed if request.seed is not None else int(torch.randint(0, 2**32 - 1, (1,)).item())
            generator = torch.Generator(device=self._device).manual_seed(seed)
            
            callback = self._create_progress_callback(task_id, request.num_inference_steps) if stream else None
            
            loop = asyncio.get_event_loop()
            
            def run_inference():
                return self._img2img_pipe(
                    prompt=request.prompt,
                    negative_prompt=request.negative_prompt or None,
                    image=init_image,
                    strength=request.strength,
                    num_inference_steps=request.num_inference_steps,
                    guidance_scale=request.guidance_scale,
                    num_images_per_prompt=request.num_images,
                    generator=generator,
                    callback_on_step_end=callback,
                )
            
            output = await loop.run_in_executor(None, run_inference)
            images_base64 = [self._encode_image_to_base64(img) for img in output.images]
            
            yield {
                "type": "result",
                "images": images_base64,
                "seed": seed,
                "task_id": task_id,
            }
            
        except Exception as e:
            uvicorn_logger.error(f"img2img error: {e}")
            yield {"type": "error", "error": str(e)}
        finally:
            del self._tasks[task_id]
            gc.collect()

    async def _img2img_sse(self, request: Img2ImgRequest) -> AsyncGenerator[str, None]:
        """图生图 SSE 生成器"""
        async for data in self._img2img(request, stream=True):
            yield f"data: {json.dumps(data)}\n\n"

    async def _inpaint(self, request: InpaintRequest, stream: bool = False) -> AsyncGenerator[Dict[str, Any], None]:
        """图像修复推理"""
        if not self._inpaint_pipe:
            yield {"type": "error", "error": "inpaint pipeline not available"}
            return
        
        torch = self._modules["torch"]
        task_id = str(uuid.uuid4())
        self._tasks[task_id] = {"progress": 0, "step": 0}
        
        try:
            init_image = self._decode_base64_image(request.image)
            mask_image = self._decode_base64_image(request.mask_image)
            
            width = request.width or init_image.width
            height = request.height or init_image.height
            
            seed = request.seed if request.seed is not None else int(torch.randint(0, 2**32 - 1, (1,)).item())
            generator = torch.Generator(device=self._device).manual_seed(seed)
            
            callback = self._create_progress_callback(task_id, request.num_inference_steps) if stream else None
            
            loop = asyncio.get_event_loop()
            
            def run_inference():
                return self._inpaint_pipe(
                    prompt=request.prompt,
                    negative_prompt=request.negative_prompt or None,
                    image=init_image,
                    mask_image=mask_image,
                    width=width,
                    height=height,
                    strength=request.strength,
                    num_inference_steps=request.num_inference_steps,
                    guidance_scale=request.guidance_scale,
                    num_images_per_prompt=request.num_images,
                    generator=generator,
                    callback_on_step_end=callback,
                )
            
            output = await loop.run_in_executor(None, run_inference)
            images_base64 = [self._encode_image_to_base64(img) for img in output.images]
            
            yield {
                "type": "result",
                "images": images_base64,
                "seed": seed,
                "task_id": task_id,
            }
            
        except Exception as e:
            uvicorn_logger.error(f"inpaint error: {e}")
            yield {"type": "error", "error": str(e)}
        finally:
            del self._tasks[task_id]
            gc.collect()

    async def _inpaint_sse(self, request: InpaintRequest) -> AsyncGenerator[str, None]:
        """图像修复 SSE 生成器"""
        async for data in self._inpaint(request, stream=True):
            yield f"data: {json.dumps(data)}\n\n"

    # ========== ieops 协议兼容 ==========
    async def _ieops_sse_generator(self, request: bytes) -> AsyncGenerator[str, None]:
        """ieops 协议 SSE 生成器"""
        async for item in self._ieops_generator(request):
            code = item["code"]
            data = item["data"]
            yield f"{json.dumps({'code': code, 'data': data}, ensure_ascii=False)}\n\n"

    async def _ieops_generator(self, request: bytes) -> AsyncGenerator[Dict[str, Any], None]:
        """ieops 协议生成器"""
        start = int(time.time() * 1000)
        
        try:
            request_data = json.loads(request.decode("utf-8"))
            headers = request_data.get("headers", {})
            payload = request_data.get("payload", request_data)
        except json.JSONDecodeError:
            payload = json.loads(request.decode("utf-8"))
            headers = {}
        
        trace_id = str(headers.get("x-trace-id", f"traceid-{await arandstr(8)}"))
        logger = log.get_logger(trace_id=trace_id)
        logger.info("Diffuser request received")
        
        if isinstance(payload, dict):
            task_type = payload.get("type", "txt2img")
            task_params = payload.get("params", payload)
        else:
            task_params = json.loads(payload) if isinstance(payload, str) else payload
            task_type = task_params.get("type", "txt2img")
        
        if self._on_request_received:
            task_params = self._on_request_received(task_params)
        
        try:
            if task_type == "txt2img":
                req = Txt2ImgRequest(**task_params)
                generator = self._txt2img(req, stream=True)
            elif task_type == "img2img":
                req = Img2ImgRequest(**task_params)
                generator = self._img2img(req, stream=True)
            elif task_type == "inpaint":
                req = InpaintRequest(**task_params)
                generator = self._inpaint(req, stream=True)
            else:
                yield {"code": SERVER_CODE_ERROR, "data": f"Unknown task type: {task_type}"}
                return
            
            async for data in generator:
                if data.get("type") == "error":
                    yield {"code": SERVER_CODE_ERROR, "data": data["error"]}
                    return
                elif data.get("type") == "progress":
                    yield {"code": SERVER_CODE_OK, "data": json.dumps(data, ensure_ascii=False)}
                elif data.get("type") == "result":
                    yield {"code": SERVER_CODE_OK, "data": json.dumps(data, ensure_ascii=False)}
                    yield {"code": SERVER_CODE_STOP, "data": ""}
                    return
            
            yield {"code": SERVER_CODE_STOP, "data": ""}
            
        except Exception as e:
            logger.error(f"Error: {e}")
            yield {"code": SERVER_CODE_ERROR, "data": str(e)}
        finally:
            cost = int(time.time() * 1000) - start
            logger.info(f"Generation completed, cost: {cost}ms")



"""
文件系统SDK - 与brslet的filesystem接口对接
支持文件上传、下载、同步等操作
"""
import os
from typing import Optional, Generator, Union, IO, Dict, Any, List
from dataclasses import dataclass

# 从client.py导入请求方法
from ..handler.client import (
    post, get, async_post, async_get,
    Payload, Response,
)


@dataclass
class TaskInfo:
    """任务信息"""
    run_id: str
    direction: str = ""
    local_path: str = ""
    remote_path: str = ""
    status: str = ""
    progress: float = 0.0
    message: str = ""
    start_time: str = ""
    end_time: str = ""


@dataclass  
class SyncResult:
    """同步结果"""
    success: bool
    md5: str = ""
    size: int = 0
    duration: int = 0
    message: str = ""


class FileSystem:
    """文件系统操作类
    
    与brslet的filesystem接口对接，支持文件上传、下载等操作。
    
    路径格式: /root_path/pid/fs_name/oid/uid/file_path
    """
    
    def __init__(
        self, 
        path: str, 
        root_path: Optional[str] = None, 
        pid: Optional[str] = None, 
        oid: Optional[str] = None, 
        uid: Optional[str] = None, 
        fs_name: Optional[str] = None
    ):
        self._path = path
        self._root_path = root_path
        self._oid = oid
        self._uid = uid
        self._fs_name = fs_name
        self._pid = pid
        # 确保本地目录存在
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

    def write(
        self, 
        file_path: str, 
        data: Union[str, bytes], 
        mode: str = 'w', 
        encoding: str = 'utf-8',
        wait: bool = True,
        timeout: int = 3600
    ) -> Union[Response, TaskInfo, Exception]:
        """写入文件并同步到远程存储
        
        Args:
            file_path: 相对于FileSystem根路径的文件路径
            data: 要写入的数据，文本模式传str，二进制模式传bytes
            mode: 'w'文本模式，'wb'二进制模式
            encoding: 文本模式的编码，二进制模式忽略此参数
            wait: 是否等待上传完成
            timeout: 等待超时时间（秒）
            
        Returns:
            如果wait=True，返回Response；否则返回TaskInfo
        """
        try:
            if mode not in ['w', 'wb']:
                raise ValueError("mode must be 'w' or 'wb'")
            
            # 写入本地文件
            full_path = os.path.join(self._path, file_path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            
            if mode == 'w':
                with open(full_path, mode, encoding=encoding) as file:
                    file.write(data)
            else:
                with open(full_path, mode) as file:
                    file.write(data)
            
            # 调用brslet上传接口
            resp = post(Payload(path="fs/upload", payload={
                "file": file_path,
                "root_path": self._root_path,
                "pid": self._pid,
                "oid": self._oid,
                "uid": self._uid,
                "fs_name": self._fs_name
            }), connector_type="brslet")
            
            if resp.message != "success":
                return resp
            
            task_info = TaskInfo(
                run_id=resp.data.get('run_id', ''),
                status=resp.data.get('status', ''),
                message=resp.data.get('message', '')
            )
            
            if wait:
                return self.wait_task(task_info.run_id, timeout=timeout)
            
            return task_info
            
        except Exception as e:
            return e

    async def write_async(
        self, 
        file_path: str, 
        data: Union[str, bytes], 
        mode: str = 'w', 
        encoding: str = 'utf-8',
        wait: bool = True,
        timeout: int = 3600
    ) -> Union[Response, TaskInfo, Exception]:
        """异步写入文件并同步到远程存储"""
        try:
            if mode not in ['w', 'wb']:
                raise ValueError("mode must be 'w' or 'wb'")
            
            # 写入本地文件
            full_path = os.path.join(self._path, file_path)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            
            if mode == 'w':
                with open(full_path, mode, encoding=encoding) as file:
                    file.write(data)
            else:
                with open(full_path, mode) as file:
                    file.write(data)
            
            # 调用brslet上传接口
            resp = await async_post(Payload(path="fs/upload", payload={
                "file": file_path,
                "root_path": self._root_path,
                "pid": self._pid,
                "oid": self._oid,
                "uid": self._uid,
                "fs_name": self._fs_name
            }), connector_type="brslet")
            
            if resp.message != "success":
                return resp
            
            task_info = TaskInfo(
                run_id=resp.data.get('run_id', ''),
                status=resp.data.get('status', ''),
                message=resp.data.get('message', '')
            )
            
            if wait:
                return await self.wait_task_async(task_info.run_id, timeout=timeout)
            
            return task_info
            
        except Exception as e:
            return e

    def read(
        self, 
        file_path: str, 
        mode: str = 'r', 
        encoding: str = 'utf-8', 
        timeout: int = 3600
    ) -> Union[IO, Exception]:
        """读取文件，如果本地不存在则从远程同步
        
        Args:
            file_path: 相对于FileSystem根路径的文件路径
            mode: 'r'文本模式，'rb'二进制模式
            encoding: 文本模式的编码，二进制模式忽略此参数
            timeout: 等待远程同步的超时时间（秒）
            
        Returns:
            打开的文件对象
        """
        try:
            if mode not in ['r', 'rb']:
                raise ValueError("mode must be 'r' or 'rb'")
            
            full_path = os.path.join(self._path, file_path)
            
            # 如果本地文件不存在，从远程下载
            if not os.path.exists(full_path):
                resp = post(Payload(path="fs/download", payload={
                    "file": file_path,
                    "root_path": self._root_path,
                    "pid": self._pid,
                    "oid": self._oid,
                    "uid": self._uid,
                    "fs_name": self._fs_name
                }), connector_type="brslet")
                
                if resp.message != "success":
                    return Exception(resp.message)
                
                # 等待下载完成
                run_id = resp.data.get('run_id', '')
                if run_id:
                    wait_resp = self.wait_task(run_id, timeout=timeout)
                    if wait_resp.message != "success":
                        return Exception(wait_resp.message)
            
            # 打开文件
            if mode == 'rb':
                return open(full_path, mode)
            else:
                return open(full_path, mode, encoding=encoding)
                
        except Exception as e:
            return e

    async def read_async(
        self, 
        file_path: str, 
        mode: str = 'r', 
        encoding: str = 'utf-8', 
        timeout: int = 3600
    ) -> Union[IO, Exception]:
        """异步读取文件，如果本地不存在则从远程同步"""
        try:
            if mode not in ['r', 'rb']:
                raise ValueError("mode must be 'r' or 'rb'")
            
            full_path = os.path.join(self._path, file_path)
            
            if not os.path.exists(full_path):
                resp = await async_post(Payload(path="fs/download", payload={
                    "file": file_path,
                    "root_path": self._root_path,
                    "pid": self._pid,
                    "oid": self._oid,
                    "uid": self._uid,
                    "fs_name": self._fs_name
                }), connector_type="brslet")
                
                if resp.message != "success":
                    return Exception(resp.message)
                
                run_id = resp.data.get('run_id', '')
                if run_id:
                    wait_resp = await self.wait_task_async(run_id, timeout=timeout)
                    if wait_resp.message != "success":
                        return Exception(wait_resp.message)
            
            if mode == 'rb':
                return open(full_path, mode)
            else:
                return open(full_path, mode, encoding=encoding)
                
        except Exception as e:
            return e

    def get_task(self, run_id: str) -> Union[TaskInfo, Exception]:
        """获取任务状态
        
        Args:
            run_id: 任务ID
            
        Returns:
            任务信息
        """
        try:
            resp = get(f"fs/task/{run_id}", connector_type="brslet")
            if resp.message != "success":
                return Exception(resp.message)
            
            data = resp.data
            return TaskInfo(
                run_id=data.get('run_id', ''),
                direction=data.get('direction', ''),
                local_path=data.get('local_path', ''),
                remote_path=data.get('remote_path', ''),
                status=data.get('status', ''),
                progress=data.get('progress', 0.0),
                message=data.get('message', ''),
                start_time=data.get('start_time', ''),
                end_time=data.get('end_time', '')
            )
        except Exception as e:
            return e

    async def get_task_async(self, run_id: str) -> Union[TaskInfo, Exception]:
        """异步获取任务状态"""
        try:
            resp = await async_get(f"fs/task/{run_id}", connector_type="brslet")
            if resp.message != "success":
                return Exception(resp.message)
            
            data = resp.data
            return TaskInfo(
                run_id=data.get('run_id', ''),
                direction=data.get('direction', ''),
                local_path=data.get('local_path', ''),
                remote_path=data.get('remote_path', ''),
                status=data.get('status', ''),
                progress=data.get('progress', 0.0),
                message=data.get('message', ''),
                start_time=data.get('start_time', ''),
                end_time=data.get('end_time', '')
            )
        except Exception as e:
            return e

    def list_tasks(self) -> Union[List[TaskInfo], Exception]:
        """获取所有任务列表"""
        try:
            resp = get("fs/tasks", connector_type="brslet")
            if resp.message != "success":
                return Exception(resp.message)
            
            tasks = []
            for item in resp.data or []:
                tasks.append(TaskInfo(
                    run_id=item.get('run_id', ''),
                    direction=item.get('direction', ''),
                    local_path=item.get('local_path', ''),
                    remote_path=item.get('remote_path', ''),
                    status=item.get('status', ''),
                    progress=item.get('progress', 0.0),
                    message=item.get('message', ''),
                    start_time=item.get('start_time', ''),
                    end_time=item.get('end_time', '')
                ))
            return tasks
        except Exception as e:
            return e

    async def list_tasks_async(self) -> Union[List[TaskInfo], Exception]:
        """异步获取所有任务列表"""
        try:
            resp = await async_get("fs/tasks", connector_type="brslet")
            if resp.message != "success":
                return Exception(resp.message)
            
            tasks = []
            for item in resp.data or []:
                tasks.append(TaskInfo(
                    run_id=item.get('run_id', ''),
                    direction=item.get('direction', ''),
                    local_path=item.get('local_path', ''),
                    remote_path=item.get('remote_path', ''),
                    status=item.get('status', ''),
                    progress=item.get('progress', 0.0),
                    message=item.get('message', ''),
                    start_time=item.get('start_time', ''),
                    end_time=item.get('end_time', '')
                ))
            return tasks
        except Exception as e:
            return e

    def cancel_task(self, run_id: str) -> Response:
        """取消任务
        
        Args:
            run_id: 任务ID
        """
        return post(Payload(path=f"fs/task/{run_id}/cancel", payload={}), connector_type="brslet")

    async def cancel_task_async(self, run_id: str) -> Response:
        """异步取消任务"""
        return await async_post(Payload(path=f"fs/task/{run_id}/cancel", payload={}), connector_type="brslet")

    def wait_task(self, run_id: str, timeout: int = 3600) -> Response:
        """等待任务完成
        
        Args:
            run_id: 任务ID
            timeout: 超时时间（秒）
            
        Returns:
            包含同步结果的Response
        """
        resp = post(Payload(path=f"fs/task/{run_id}/wait", payload={"timeout": timeout}), 
                    timeout=float(timeout + 10), connector_type="brslet")
        if resp.message != "success":
            return resp
        
        data = resp.data or {}
        return Response(
            message=data.get('message', 'success'),
            data=SyncResult(
                success=data.get('success', False),
                md5=data.get('md5', ''),
                size=data.get('size', 0),
                duration=data.get('duration', 0),
                message=data.get('message', '')
            )
        )

    async def wait_task_async(self, run_id: str, timeout: int = 3600) -> Response:
        """异步等待任务完成"""
        resp = await async_post(Payload(path=f"fs/task/{run_id}/wait", payload={"timeout": timeout}), 
                                timeout=float(timeout + 10), connector_type="brslet")
        if resp.message != "success":
            return resp
        
        data = resp.data or {}
        return Response(
            message=data.get('message', 'success'),
            data=SyncResult(
                success=data.get('success', False),
                md5=data.get('md5', ''),
                size=data.get('size', 0),
                duration=data.get('duration', 0),
                message=data.get('message', '')
            )
        )

    def walk(self, path: str = "") -> Generator[str, None, None]:
        """遍历文件系统目录
        
        Args:
            path: 相对于FileSystem根路径的目录路径
            
        Yields:
            每个文件和子目录的完整路径
        """
        try:
            resp = post(Payload(path="fs/list", payload={
                "root_path": self._root_path,
                "path": path,
                "pid": self._pid,
                "oid": self._oid,
                "uid": self._uid,
            }), connector_type="brslet")
            
            if resp.message != "success":
                return
            
            items = resp.data.get('items', []) if resp.data else []
            for item in items:
                item_path = os.path.join(path, item['name'])
                yield item_path
                if item.get('type') == 'dir':
                    yield from self.walk(item_path)
                    
        except Exception:
            return

    def exists(self, file_path: str) -> bool:
        """检查本地文件是否存在"""
        full_path = os.path.join(self._path, file_path)
        return os.path.exists(full_path)

    def delete_local(self, file_path: str) -> bool:
        """删除本地文件
        
        Args:
            file_path: 相对路径
            
        Returns:
            是否删除成功
        """
        try:
            full_path = os.path.join(self._path, file_path)
            if os.path.exists(full_path):
                os.remove(full_path)
                return True
            return False
        except Exception:
            return False

    def get_local_path(self, file_path: str) -> str:
        """获取文件的本地完整路径"""
        return os.path.join(self._path, file_path)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def Open(
    root_path: Optional[str] = None, 
    pid: Optional[str] = None, 
    oid: Optional[str] = None, 
    uid: Optional[str] = None, 
    fs_name: Optional[str] = None
) -> FileSystem:
    """打开文件系统
    
    Args:
        root_path: 根路径，默认从环境变量IEOPS_STORAGE_ROOT_PATH获取
        pid: 项目ID，默认从环境变量IEOPS_STORAGE_PID获取
        oid: 组织ID，默认从环境变量IEOPS_STORAGE_OID获取
        uid: 用户ID，默认从环境变量IEOPS_STORAGE_UID获取
        fs_name: 文件系统名称，默认从环境变量IEOPS_STORAGE_FS_NAME获取
        
    Returns:
        FileSystem实例
    """
    if fs_name is None or fs_name == '':
        fs_name = os.getenv('IEOPS_STORAGE_FS_NAME', 'jfs-dev')
    if pid is None or pid == '':
        pid = os.getenv('IEOPS_STORAGE_PID', 'u-00000000')
    if root_path is None or root_path == '':
        root_path = os.getenv('IEOPS_STORAGE_ROOT_PATH', "/mnt/juicefs")
    if oid is None or oid == '':
        oid = os.getenv('IEOPS_STORAGE_OID', 'ieops')
    if uid is None or uid == '':
        uid = os.getenv('IEOPS_STORAGE_UID', 'ieops')
    
    if not os.path.exists(root_path):
        os.makedirs(root_path, exist_ok=True)
    
    # 构建完整路径: /root_path/pid/fs_name/oid/uid
    path = os.path.join(root_path, pid, fs_name, oid, uid)
    
    return FileSystem(
        path=path, 
        root_path=root_path,
        pid=pid, 
        oid=oid, 
        uid=uid, 
        fs_name=fs_name
    )


# 便捷函数
def upload(
    file_path: str,
    data: Union[str, bytes],
    mode: str = 'w',
    encoding: str = 'utf-8',
    wait: bool = True,
    timeout: int = 3600,
    **kwargs
) -> Union[Response, TaskInfo, Exception]:
    """便捷上传函数
    
    Args:
        file_path: 文件路径
        data: 文件内容
        mode: 写入模式 'w' 或 'wb'
        encoding: 编码
        wait: 是否等待完成
        timeout: 超时时间
        **kwargs: 传递给Open()的参数
    """
    with Open(**kwargs) as fs:
        return fs.write(file_path, data, mode, encoding, wait, timeout)


async def upload_async(
    file_path: str,
    data: Union[str, bytes],
    mode: str = 'w',
    encoding: str = 'utf-8',
    wait: bool = True,
    timeout: int = 3600,
    **kwargs
) -> Union[Response, TaskInfo, Exception]:
    """异步便捷上传函数"""
    fs = Open(**kwargs)
    return await fs.write_async(file_path, data, mode, encoding, wait, timeout)


def download(
    file_path: str,
    mode: str = 'r',
    encoding: str = 'utf-8',
    timeout: int = 3600,
    **kwargs
) -> Union[IO, Exception]:
    """便捷下载函数
    
    Args:
        file_path: 文件路径
        mode: 读取模式 'r' 或 'rb'
        encoding: 编码
        timeout: 超时时间
        **kwargs: 传递给Open()的参数
    """
    fs = Open(**kwargs)
    return fs.read(file_path, mode, encoding, timeout)


async def download_async(
    file_path: str,
    mode: str = 'r',
    encoding: str = 'utf-8',
    timeout: int = 3600,
    **kwargs
) -> Union[IO, Exception]:
    """异步便捷下载函数"""
    fs = Open(**kwargs)
    return await fs.read_async(file_path, mode, encoding, timeout)


__all__ = [
    'FileSystem',
    'TaskInfo', 
    'SyncResult',
    'Response',
    'Open',
    'upload',
    'upload_async',
    'download',
    'download_async',
]



# /root/code/python/ieops-python/sdk/src/utils/subprocess_manager.py
"""
通用子进程管理器

提供子进程的启动、停止、监控、日志转发等功能，
可被各种 Server 复用（如 ComfyUIServer、vLLM Server 等）。

使用示例:
    manager = SubprocessManager(
        start_cmd="python main.py --port 8188",
        work_dir="/app/comfyui",
        health_check=lambda: check_http("http://localhost:8188/health"),
        startup_timeout=120,
        exit_on_failure=True,
    )
    
    # 启动子进程
    await manager.start()
    
    # 检查状态
    if manager.is_running:
        print("Process is running")
    
    # 停止子进程
    await manager.stop()
"""

import asyncio
import atexit
import os
import signal
import subprocess
import sys
import time
from typing import Any, Callable, Coroutine, Dict, List, Optional

from .log import uvicorn_logger

# 全局注册的子进程管理器列表，用于 atexit 清理
_registered_managers: List["SubprocessManager"] = []


def _cleanup_all_subprocesses():
    """atexit 回调：清理所有注册的子进程"""
    for manager in _registered_managers:
        manager._sync_kill()


# 注册 atexit 清理函数
atexit.register(_cleanup_all_subprocesses)


class SubprocessManager:
    """
    通用子进程管理器
    
    功能：
    - 启动子进程并等待就绪
    - 转发子进程日志
    - 后台监控子进程状态
    - 优雅停止子进程
    - 异常退出时可选择终止主进程
    """
    
    def __init__(
        self,
        start_cmd: str,
        work_dir: str,
        *,
        name: str = "subprocess",
        extra_env: Optional[Dict[str, str]] = None,
        health_check: Optional[Callable[[], Coroutine[Any, Any, bool]]] = None,
        startup_timeout: int = 120,
        health_check_interval: float = 1.0,
        monitor_interval: float = 5.0,
        graceful_shutdown_timeout: float = 5.0,
        exit_on_failure: bool = True,
        log_prefix: Optional[str] = None,
    ) -> None:
        """
        Args:
            start_cmd: 启动命令，如 "python main.py --port 8188"
            work_dir: 工作目录
            name: 进程名称（用于日志）
            extra_env: 额外的环境变量
            health_check: 健康检查函数，返回 True 表示就绪
            startup_timeout: 启动超时时间（秒）
            health_check_interval: 健康检查间隔（秒）
            monitor_interval: 后台监控间隔（秒）
            graceful_shutdown_timeout: 优雅关闭超时（秒）
            exit_on_failure: 子进程异常退出时是否调用 sys.exit(1)
            log_prefix: 日志前缀，默认使用 name
        """
        self._start_cmd = start_cmd
        self._work_dir = work_dir
        self._name = name
        self._extra_env = extra_env or {}
        self._health_check = health_check
        self._startup_timeout = startup_timeout
        self._health_check_interval = health_check_interval
        self._monitor_interval = monitor_interval
        self._graceful_shutdown_timeout = graceful_shutdown_timeout
        self._exit_on_failure = exit_on_failure
        self._log_prefix = log_prefix or f"[{name}]"
        
        self._process: Optional[subprocess.Popen] = None
        self._is_ready = False
        self._shutting_down = False
        self._monitor_task: Optional[asyncio.Task] = None
        self._log_task: Optional[asyncio.Task] = None

    @property
    def is_running(self) -> bool:
        """检查子进程是否正在运行"""
        return self._process is not None and self._process.poll() is None

    @property
    def is_ready(self) -> bool:
        """检查子进程是否已就绪（通过健康检查）"""
        return self._is_ready

    @property
    def exit_code(self) -> Optional[int]:
        """获取子进程退出码，如果还在运行返回 None"""
        if self._process is None:
            return None
        return self._process.poll()

    @property
    def pid(self) -> Optional[int]:
        """获取子进程 PID"""
        if self._process is None:
            return None
        return self._process.pid

    async def start(self) -> None:
        """
        启动子进程
        
        1. 启动进程
        2. 启动日志转发
        3. 等待健康检查通过
        4. 启动后台监控
        
        Raises:
            RuntimeError: 进程启动失败或在启动过程中退出
            TimeoutError: 启动超时
        """
        uvicorn_logger.info(f"{self._log_prefix} Starting: {self._start_cmd}")
        uvicorn_logger.info(f"{self._log_prefix} Work directory: {self._work_dir}")
        
        # 准备环境变量
        env = os.environ.copy()
        env.update(self._extra_env)
        
        # 启动子进程
        self._process = subprocess.Popen(
            self._start_cmd,
            shell=True,
            cwd=self._work_dir,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            preexec_fn=self._setup_child_process,  # 设置子进程属性
        )
        
        uvicorn_logger.info(f"{self._log_prefix} Started with PID: {self._process.pid}")
        
        # 注册到全局管理器列表，用于 atexit 清理
        _registered_managers.append(self)
        
        # 启动日志转发
        self._log_task = asyncio.create_task(self._forward_logs())
        
        # 等待健康检查通过
        await self._wait_for_ready()
        
        # 启动后台监控
        self._monitor_task = asyncio.create_task(self._monitor_process())
    
    def _setup_child_process(self) -> None:
        """
        子进程启动前的设置（在 fork 后、exec 前调用）
        
        1. 创建新的进程组（方便后续 kill 整个组）
        2. 设置 prctl PR_SET_PDEATHSIG（Linux 特性：父进程死亡时子进程收到信号）
        """
        # 创建新的进程组
        os.setsid()
        
        # Linux 特性：设置父进程死亡时子进程收到的信号
        try:
            import ctypes
            libc = ctypes.CDLL("libc.so.6", use_errno=True)
            PR_SET_PDEATHSIG = 1
            # 当父进程死亡时，子进程收到 SIGKILL
            libc.prctl(PR_SET_PDEATHSIG, signal.SIGKILL)
        except Exception:
            pass  # 非 Linux 系统或没有 prctl 支持，忽略

    async def stop(self) -> None:
        """
        优雅停止子进程
        
        1. 发送 SIGTERM
        2. 等待进程退出
        3. 超时后发送 SIGKILL
        """
        self._shutting_down = True
        
        # 取消监控任务
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
        
        if not self._process:
            return
        
        uvicorn_logger.info(f"{self._log_prefix} Stopping process...")
        
        try:
            # 先尝试 SIGTERM
            os.killpg(os.getpgid(self._process.pid), signal.SIGTERM)
            
            # 等待进程退出
            wait_count = int(self._graceful_shutdown_timeout / 0.5)
            for _ in range(wait_count):
                if self._process.poll() is not None:
                    uvicorn_logger.info(f"{self._log_prefix} Process stopped gracefully")
                    return
                await asyncio.sleep(0.5)
            
            # 强制 SIGKILL
            uvicorn_logger.warning(f"{self._log_prefix} Force killing process...")
            os.killpg(os.getpgid(self._process.pid), signal.SIGKILL)
            self._process.wait()
            uvicorn_logger.info(f"{self._log_prefix} Process killed")
            
        except ProcessLookupError:
            pass  # 进程已经退出
        except Exception as e:
            uvicorn_logger.error(f"{self._log_prefix} Error stopping process: {e}")
        finally:
            self._process = None
            self._is_ready = False
            # 从全局列表中移除
            if self in _registered_managers:
                _registered_managers.remove(self)
    
    def _sync_kill(self) -> None:
        """
        同步强制杀死子进程（用于 atexit 等无法使用 async 的场景）
        """
        if not self._process:
            return
        
        try:
            # 直接发送 SIGKILL 到整个进程组
            os.killpg(os.getpgid(self._process.pid), signal.SIGKILL)
            self._process.wait(timeout=2)
        except ProcessLookupError:
            pass  # 进程已经退出
        except Exception:
            pass  # 忽略错误，这是清理阶段
        finally:
            self._process = None

    async def _forward_logs(self) -> None:
        """转发子进程的日志输出"""
        if not self._process or not self._process.stdout:
            return
        
        loop = asyncio.get_event_loop()
        
        def read_line():
            if self._process and self._process.stdout:
                return self._process.stdout.readline()
            return b""
        
        while self._process and self._process.poll() is None:
            try:
                line = await loop.run_in_executor(None, read_line)
                if line:
                    log_line = line.decode("utf-8", errors="ignore").strip()
                    uvicorn_logger.info(f"{self._log_prefix} {log_line}")
            except Exception:
                break

    async def _wait_for_ready(self) -> None:
        """等待子进程就绪"""
        if not self._health_check:
            # 没有健康检查，直接认为就绪
            self._is_ready = True
            uvicorn_logger.info(f"{self._log_prefix} No health check configured, assuming ready")
            return
        
        uvicorn_logger.info(f"{self._log_prefix} Waiting for process to be ready...")
        
        start_time = time.time()
        
        while time.time() - start_time < self._startup_timeout:
            try:
                if await self._health_check():
                    self._is_ready = True
                    uvicorn_logger.info(f"{self._log_prefix} Process is ready!")
                    return
            except Exception:
                pass
            
            # 检查进程是否还活着
            if self._process and self._process.poll() is not None:
                exit_code = self._process.returncode
                uvicorn_logger.error(
                    f"{self._log_prefix} Process died during startup with exit code {exit_code}"
                )
                if self._exit_on_failure:
                    sys.exit(1)
                raise RuntimeError(f"Process died with exit code {exit_code}")
            
            await asyncio.sleep(self._health_check_interval)
        
        # 启动超时
        uvicorn_logger.error(
            f"{self._log_prefix} Process failed to start within {self._startup_timeout}s"
        )
        if self._exit_on_failure:
            sys.exit(1)
        raise TimeoutError(f"Process failed to start within {self._startup_timeout}s")

    async def _monitor_process(self) -> None:
        """后台监控子进程状态，异常退出时处理"""
        while True:
            await asyncio.sleep(self._monitor_interval)
            
            # 如果正在正常关闭，停止监控
            if self._shutting_down:
                break
            
            if self._process is None:
                break
            
            exit_code = self._process.poll()
            if exit_code is not None:
                # 进程已退出且不是正常关闭
                uvicorn_logger.error(
                    f"{self._log_prefix} Process exited unexpectedly with code {exit_code}"
                )
                if self._exit_on_failure:
                    uvicorn_logger.error(f"{self._log_prefix} Shutting down main process...")
                    sys.exit(1)
                break


async def http_health_check(
    url: str,
    timeout: float = 5.0,
    expected_status: int = 200
) -> bool:
    """
    通用 HTTP 健康检查函数
    
    Args:
        url: 健康检查 URL
        timeout: 超时时间（秒）
        expected_status: 期望的状态码
    
    Returns:
        True 如果健康检查通过
    """
    import aiohttp
    
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url,
                timeout=aiohttp.ClientTimeout(total=timeout)
            ) as resp:
                return resp.status == expected_status
    except Exception:
        return False


def create_http_health_check(
    url: str,
    timeout: float = 5.0,
    expected_status: int = 200
) -> Callable[[], Coroutine[Any, Any, bool]]:
    """
    创建 HTTP 健康检查函数的工厂方法
    
    使用示例:
        health_check = create_http_health_check("http://localhost:8188/system_stats")
        manager = SubprocessManager(
            start_cmd="...",
            work_dir="...",
            health_check=health_check,
        )
    """
    async def check() -> bool:
        return await http_health_check(url, timeout, expected_status)
    return check




# /root/code/python/ieops-python/sdk/src/handler/server/comfyui.py

import asyncio
import json
import os
import base64
import uuid
import time
from typing import Any, AsyncGenerator, Dict, List, Optional, Callable
from typing_extensions import TypedDict

import aiohttp
import uvicorn
from fastapi import FastAPI, Request, Response, WebSocket, WebSocketDisconnect
from fastapi.responses import StreamingResponse, JSONResponse
from starlette.websockets import WebSocketState
from contextlib import asynccontextmanager

from .server import BaseServer
from ...utils.log import *
from ...utils.util import APP_ID, arandstr
from ...utils.subprocess_manager import SubprocessManager, create_http_health_check

# ========== 环境变量配置 ==========
_MODEL_SERVER_SOCKET = os.getenv("MODEL_SERVER_SOCKET", f"{APP_ID}.sock")
_MODEL_SERVER_PORT = int(os.getenv("MODEL_SERVER_PORT", 8001))
_MODEL_SERVER_HOST = os.getenv("MODEL_SERVER_HOST", "127.0.0.1")
_GRACEFUL_SHUTDOWN_TIME = int(os.getenv("GRACEFUL_SHUTDOWN_TIME", 3))
_MODEL_TIMEOUT = int(os.getenv("MODEL_TIMEOUT", 600))

# ComfyUI 专用配置
_COMFYUI_PORT = int(os.getenv("COMFYUI_PORT", 8188))
_COMFYUI_HOST = os.getenv("COMFYUI_HOST", "127.0.0.1")
_COMFYUI_START_CMD = os.getenv("COMFYUI_START_CMD", "python main.py")
_COMFYUI_WORK_DIR = os.getenv("COMFYUI_WORK_DIR", "/workspace/ComfyUI")
_COMFYUI_STARTUP_TIMEOUT = int(os.getenv("COMFYUI_STARTUP_TIMEOUT", 120))

SERVER_CODE_OK = 0
SERVER_CODE_ERROR = 1
SERVER_CODE_STOP = 2


class ComfyUITask(TypedDict):
    prompt_id: str
    trace_id: str
    status: str
    progress: float
    images: List[str]


class ComfyUIServer(BaseServer):
    """
    ComfyUI 代理服务器 - 管理独立 ComfyUI 进程的生命周期
    
    架构：
    ┌─────────────────┐     ┌──────────────────┐     ┌────────────────┐
    │  外部客户端      │────▶│  ComfyUIServer   │────▶│  ComfyUI 进程  │
    │  (HTTP/SSE)     │     │  (FastAPI 代理)   │     │  (子进程)      │
    └─────────────────┘     └──────────────────┘     └────────────────┘
    """
    
    def __init__(
        self,
        start_cmd: Optional[str] = None,
        work_dir: Optional[str] = None,
        comfyui_host: Optional[str] = None,
        comfyui_port: Optional[int] = None,
        on_workflow_received: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        extra_env: Optional[Dict[str, str]] = None,
        startup_timeout: Optional[int] = None,
    ) -> None:
        """
        Args:
            start_cmd: 启动 ComfyUI 的命令，如 "python main.py --port 8188"
            work_dir: ComfyUI 工作目录
            comfyui_host: ComfyUI 监听地址
            comfyui_port: ComfyUI 监听端口
            on_workflow_received: 工作流预处理回调
            extra_env: 额外的环境变量
            startup_timeout: 启动超时时间（秒）
        """
        self._comfyui_host = comfyui_host or _COMFYUI_HOST
        self._comfyui_port = comfyui_port or _COMFYUI_PORT
        self._comfyui_url = f"http://{self._comfyui_host}:{self._comfyui_port}"
        self._on_workflow_received = on_workflow_received
        
        # 构建启动命令（自动添加端口和监听地址参数）
        cmd = start_cmd or _COMFYUI_START_CMD
        if "--port" not in cmd:
            cmd += f" --port {self._comfyui_port}"
        if "--listen" not in cmd:
            cmd += f" --listen {self._comfyui_host}"
        
        # 使用通用子进程管理器
        # 注意：使用 /queue 而不是 /system_stats 做健康检查，因为某些版本的 ComfyUI 
        # 在 /system_stats 中有 bug（缺少 import sys）
        self._process_manager = SubprocessManager(
            start_cmd=cmd,
            work_dir=work_dir or _COMFYUI_WORK_DIR,
            name="ComfyUI",
            extra_env=extra_env,
            health_check=create_http_health_check(f"{self._comfyui_url}/queue"),
            startup_timeout=startup_timeout or _COMFYUI_STARTUP_TIMEOUT,
            exit_on_failure=True,
        )
        
        self._tasks: Dict[str, ComfyUITask] = {}
        self._http_session: Optional[aiohttp.ClientSession] = None
        self._app = FastAPI(lifespan=self._lifespan)
        self._server: Optional[uvicorn.Server] = None
        
        self._setup_routes()

    @asynccontextmanager
    async def _lifespan(self, app: FastAPI):
        """FastAPI 生命周期管理"""
        uvicorn_logger.info("ComfyUI Server starting...")
        
        # 1. 启动 ComfyUI 子进程（使用通用子进程管理器）
        await self._process_manager.start()
        
        # 2. 初始化 HTTP 客户端
        self._http_session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=_MODEL_TIMEOUT)
        )
        
        yield
        
        # 清理
        uvicorn_logger.info("ComfyUI Server shutting down...")
        
        # 关闭 HTTP 客户端
        if self._http_session:
            await self._http_session.close()
        
        # 停止 ComfyUI 进程
        await self._process_manager.stop()
        
        # 清理 socket 文件
        if _MODEL_SERVER_SOCKET and os.path.exists(_MODEL_SERVER_SOCKET):
            os.remove(_MODEL_SERVER_SOCKET)

    def shutdown(self) -> None:
        """请求服务器优雅关闭"""
        if self._server:
            self._server.should_exit = True

    async def serve(self) -> None:
        """启动代理服务器"""
        if _MODEL_SERVER_SOCKET:
            config = uvicorn.Config(
                self._app,
                uds=_MODEL_SERVER_SOCKET,
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=_GRACEFUL_SHUTDOWN_TIME
            )
        else:
            config = uvicorn.Config(
                self._app,
                host=_MODEL_SERVER_HOST,
                port=_MODEL_SERVER_PORT,
                log_level="info",
                loop="asyncio",
                timeout_graceful_shutdown=_GRACEFUL_SHUTDOWN_TIME
            )
        
        self._server = uvicorn.Server(config)
        await self._server.serve()

    def _setup_routes(self) -> None:
        """
        设置 API 路由 - 与 ComfyUI 原生 API 路径保持一致
        
        ComfyUI 原生 API:
        - GET  /                 - 前端页面
        - GET  /system_stats     - 系统状态
        - GET  /queue            - 队列信息
        - POST /prompt           - 提交工作流
        - POST /interrupt        - 中断生成
        - GET  /history          - 历史记录
        - GET  /history/{prompt_id} - 指定历史
        - GET  /view             - 查看图像
        - POST /upload/image     - 上传图像
        - GET  /embeddings       - embeddings列表
        - GET  /extensions       - 扩展列表
        - GET  /object_info      - 节点信息
        - WS   /ws               - WebSocket
        """
        
        # ========== 健康检查（额外添加）==========
        @self._app.get("/health")
        async def health():
            """健康检查接口"""
            if not self._process_manager.is_ready:
                return JSONResponse({"status": "initializing"}, status_code=503)
            
            if not self._process_manager.is_running:
                return JSONResponse(
                    {"status": "error", "message": "ComfyUI process died"},
                    status_code=503
                )
            return {"status": "ok"}

        # ========== ComfyUI 原生 API - 直接代理 ==========
        
        @self._app.get("/")
        async def root():
            """代理 ComfyUI 根路径（前端页面）"""
            return await self._proxy_request("GET", "/")

        @self._app.get("/system_stats")
        async def system_stats():
            """系统状态"""
            return await self._proxy_request("GET", "/system_stats")

        @self._app.get("/queue")
        async def get_queue():
            """获取队列信息"""
            return await self._proxy_request("GET", "/queue")

        @self._app.post("/queue")
        async def post_queue(request: Request):
            """队列操作（清空等）"""
            body = await request.body()
            return await self._proxy_request("POST", "/queue", body=body)

        @self._app.post("/prompt")
        async def prompt(request: Request):
            """
            提交工作流 - 核心接口
            支持两种模式:
            1. 直接代理到 ComfyUI（默认）
            2. 带进度流式返回（通过 X-Stream-Response: true 头开启）
            """
            body = await request.body()
            
            # 检查是否需要流式响应
            if request.headers.get("X-Stream-Response", "").lower() == "true":
                return StreamingResponse(
                    self._sse_generator(body),
                    media_type="text/event-stream",
                    headers={
                        "Cache-Control": "no-cache",
                        "Connection": "keep-alive",
                        "X-Accel-Buffering": "no",
                    }
                )
            
            # 默认直接代理
            return await self._proxy_request("POST", "/prompt", body=body)

        @self._app.post("/interrupt")
        async def interrupt():
            """中断当前生成"""
            return await self._proxy_request("POST", "/interrupt")

        @self._app.get("/history")
        async def get_history():
            """获取所有历史记录"""
            return await self._proxy_request("GET", "/history")

        @self._app.get("/history/{prompt_id}")
        async def get_history_by_id(prompt_id: str):
            """获取指定 prompt 的历史"""
            return await self._proxy_request("GET", f"/history/{prompt_id}")

        @self._app.post("/history")
        async def post_history(request: Request):
            """清除历史"""
            body = await request.body()
            return await self._proxy_request("POST", "/history", body=body)

        @self._app.get("/view")
        async def view(request: Request):
            """查看/下载图像"""
            return await self._proxy_request(
                "GET", "/view", 
                params=dict(request.query_params)
            )

        @self._app.post("/upload/image")
        async def upload_image(request: Request):
            """上传图像"""
            body = await request.body()
            headers = {k: v for k, v in request.headers.items() 
                      if k.lower() in ("content-type", "content-length")}
            return await self._proxy_request("POST", "/upload/image", body=body, headers=headers)

        @self._app.post("/upload/mask")
        async def upload_mask(request: Request):
            """上传遮罩"""
            body = await request.body()
            headers = {k: v for k, v in request.headers.items() 
                      if k.lower() in ("content-type", "content-length")}
            return await self._proxy_request("POST", "/upload/mask", body=body, headers=headers)

        @self._app.get("/embeddings")
        async def get_embeddings():
            """获取 embeddings 列表"""
            return await self._proxy_request("GET", "/embeddings")

        @self._app.get("/extensions")
        async def get_extensions():
            """获取扩展列表"""
            return await self._proxy_request("GET", "/extensions")

        @self._app.get("/object_info")
        async def get_object_info():
            """获取所有节点信息"""
            return await self._proxy_request("GET", "/object_info")

        @self._app.get("/object_info/{node_class}")
        async def get_object_info_by_class(node_class: str):
            """获取指定节点类型的信息"""
            return await self._proxy_request("GET", f"/object_info/{node_class}")

        # ========== 扩展接口 - 带进度的流式生成 ==========
        
        @self._app.post("/prompt/stream")
        async def prompt_stream(request: Request):
            """
            流式提交工作流 - 扩展接口
            返回 SSE 流，实时推送进度和结果
            """
            body = await request.body()
            return StreamingResponse(
                self._sse_generator(body),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        @self._app.post("/prompt/sync")
        async def prompt_sync(request: Request):
            """
            同步提交工作流 - 扩展接口
            等待生成完成后返回所有图像（base64）
            """
            body = await request.body()
            results: List[Any] = []
            last_code = SERVER_CODE_OK
            
            async for item in self._generate(body):
                code = item["code"]
                data = item["data"]
                if code == SERVER_CODE_ERROR:
                    last_code = code
                    results.append({"error": data})
                    break
                if code == SERVER_CODE_STOP:
                    break
                results.append(data)
            
            return {"code": last_code, "results": results}

        # ========== WebSocket 代理 ==========
        @self._app.websocket("/ws")
        async def websocket_proxy(websocket: WebSocket):
            """代理 WebSocket 连接到 ComfyUI"""
            await self._proxy_websocket(websocket)
        
        # ========== 通配符路由 - 代理其他未明确定义的路径 ==========
        @self._app.api_route(
            "/{path:path}",
            methods=["GET", "POST", "PUT", "DELETE"]
        )
        async def proxy_fallback(request: Request, path: str):
            """透传其他未定义的路由到 ComfyUI"""
            body = await request.body() if request.method in ["POST", "PUT"] else None
            return await self._proxy_request(
                request.method, 
                f"/{path}", 
                body=body,
                params=dict(request.query_params)
            )

    async def _proxy_request(
        self, 
        method: str, 
        path: str, 
        body: Optional[bytes] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None
    ) -> Response:
        """通用代理请求方法"""
        url = f"{self._comfyui_url}{path}"
        
        try:
            async with self._http_session.request(
                method=method,
                url=url,
                data=body,
                params=params,
                headers=headers,
            ) as resp:
                content = await resp.read()
                
                # 过滤掉一些不应该透传的响应头
                response_headers = {}
                for k, v in resp.headers.items():
                    if k.lower() not in ("transfer-encoding", "content-encoding", "content-length"):
                        response_headers[k] = v
                
                return Response(
                    content=content,
                    status_code=resp.status,
                    headers=response_headers,
                    media_type=resp.content_type
                )
        except Exception as e:
            uvicorn_logger.error(f"Proxy error: {e}")
            return JSONResponse({"error": str(e)}, status_code=502)

    async def _proxy_websocket(self, websocket: WebSocket) -> None:
        """代理 WebSocket 连接到 ComfyUI 后端"""
        # 获取查询参数（如 clientId）
        query_string = str(websocket.scope.get("query_string", b""), "utf-8")
        ws_url = f"ws://{self._comfyui_host}:{self._comfyui_port}/ws"
        if query_string:
            ws_url += f"?{query_string}"
        
        uvicorn_logger.info(f"WebSocket proxy connecting to: {ws_url}")
        
        # 用于协调两个方向的转发任务
        should_stop = asyncio.Event()
        
        try:
            # 接受来自客户端的 WebSocket 连接
            await websocket.accept()
            
            # 创建独立的 session 用于 WebSocket，设置心跳保持连接
            async with aiohttp.ClientSession() as ws_session:
                # 连接到 ComfyUI 后端 WebSocket，设置心跳间隔
                # 压缩扩展协商现在由整个链路正确传递（Gateway <-> Proxy <-> Backend）
                async with ws_session.ws_connect(
                    ws_url,
                    heartbeat=30.0,  # 30秒心跳保持连接
                    receive_timeout=None,  # 无接收超时
                ) as backend_ws:
                    uvicorn_logger.info("WebSocket proxy connected to ComfyUI")
                    
                    async def forward_to_backend():
                        """将客户端消息转发到后端"""
                        try:
                            while not should_stop.is_set():
                                try:
                                    data = await asyncio.wait_for(
                                        websocket.receive(),
                                        timeout=5.0  # 增加超时时间
                                    )
                                    uvicorn_logger.debug(f"Client -> Backend: type={data.get('type')}")
                                    if data["type"] == "websocket.receive":
                                        if "text" in data:
                                            await backend_ws.send_str(data["text"])
                                        elif "bytes" in data:
                                            await backend_ws.send_bytes(data["bytes"])
                                    elif data["type"] == "websocket.disconnect":
                                        uvicorn_logger.info("Client disconnected")
                                        should_stop.set()
                                        break
                                except asyncio.TimeoutError:
                                    # 只是超时，继续等待
                                    continue
                        except WebSocketDisconnect:
                            uvicorn_logger.info("Client WebSocket disconnected")
                            should_stop.set()
                        except Exception as e:
                            uvicorn_logger.error(f"Forward to backend error: {e}")
                            should_stop.set()
                        finally:
                            uvicorn_logger.info("forward_to_backend task ended")
                    
                    async def forward_to_client():
                        """将后端消息转发到客户端"""
                        try:
                            while not should_stop.is_set():
                                try:
                                    msg = await asyncio.wait_for(
                                        backend_ws.receive(),
                                        timeout=5.0  # 增加超时时间
                                    )
                                    uvicorn_logger.debug(f"Backend -> Client: type={msg.type}")
                                    if websocket.client_state != WebSocketState.CONNECTED:
                                        uvicorn_logger.info("Client state not connected")
                                        should_stop.set()
                                        break
                                    if msg.type == aiohttp.WSMsgType.TEXT:
                                        await websocket.send_text(msg.data)
                                    elif msg.type == aiohttp.WSMsgType.BINARY:
                                        await websocket.send_bytes(msg.data)
                                    elif msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR):
                                        uvicorn_logger.info(f"Backend WebSocket closed: {msg.type}")
                                        should_stop.set()
                                        break
                                except asyncio.TimeoutError:
                                    # 只是超时，继续等待
                                    continue
                        except Exception as e:
                            uvicorn_logger.error(f"Forward to client error: {e}")
                            should_stop.set()
                        finally:
                            uvicorn_logger.info("forward_to_client task ended")
                    
                    # 并行运行双向转发，使用 wait 而不是 gather 以便更好地处理取消
                    tasks = [
                        asyncio.create_task(forward_to_backend()),
                        asyncio.create_task(forward_to_client()),
                    ]
                    
                    try:
                        # 等待任一任务完成
                        done, pending = await asyncio.wait(
                            tasks,
                            return_when=asyncio.FIRST_COMPLETED
                        )
                        
                        # 设置停止标志
                        should_stop.set()
                        
                        # 取消并等待剩余任务
                        for task in pending:
                            task.cancel()
                            try:
                                await task
                            except asyncio.CancelledError:
                                pass
                    except Exception as e:
                        uvicorn_logger.debug(f"WebSocket proxy wait error: {e}")
                        should_stop.set()
                        for task in tasks:
                            task.cancel()
                
        except aiohttp.ClientError as e:
            uvicorn_logger.error(f"WebSocket proxy connection error: {e}")
            if websocket.client_state == WebSocketState.CONNECTED:
                await websocket.close(code=1011, reason="Backend connection failed")
        except Exception as e:
            uvicorn_logger.error(f"WebSocket proxy error: {e}")
            if websocket.client_state == WebSocketState.CONNECTED:
                await websocket.close(code=1011, reason=f"Proxy error: {e}")
        finally:
            uvicorn_logger.info("WebSocket proxy connection closed")

    async def _sse_generator(self, request: bytes) -> AsyncGenerator[str, None]:
        """SSE 响应生成器"""
        async for item in self._generate(request):
            code = item["code"]
            data = item["data"]
            yield f"{json.dumps({'code': code, 'data': data}, ensure_ascii=False)}\n\n"

    async def _generate(self, request: bytes) -> AsyncGenerator[Dict[str, Any], None]:
        """核心生成逻辑"""
        start = int(time.time() * 1000)
        
        # 解析请求
        try:
            request_data = json.loads(request.decode("utf-8"))
            headers = request_data.get("headers", {})
            payload = request_data.get("payload", request_data)
        except json.JSONDecodeError:
            payload = json.loads(request.decode("utf-8"))
            headers = {}
        
        trace_id: str = str(headers.get("x-trace-id", "traceid-" + await arandstr(8)))
        logger = log.get_logger(trace_id=trace_id)
        logger.info("ComfyUI request received")
        
        workflow: Dict[str, Any] = payload if isinstance(payload, dict) else json.loads(payload)
        
        # 工作流预处理
        if self._on_workflow_received:
            workflow = self._on_workflow_received(workflow)
        
        client_id = str(uuid.uuid4())
        
        try:
            # 提交工作流
            prompt_id = await self._queue_prompt(workflow, client_id)
            if not prompt_id:
                yield {"code": SERVER_CODE_ERROR, "data": "Failed to queue prompt"}
                return
            
            logger.info(f"Prompt queued: {prompt_id}")
            
            # 记录任务
            self._tasks[prompt_id] = ComfyUITask(
                prompt_id=prompt_id,
                trace_id=trace_id,
                status="pending",
                progress=0.0,
                images=[]
            )
            
            # WebSocket 监听进度
            ws_url = f"ws://{self._comfyui_host}:{self._comfyui_port}/ws?clientId={client_id}"
            
            async with self._http_session.ws_connect(ws_url) as ws:
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        data = json.loads(msg.data)
                        msg_type = data.get("type")
                        msg_data = data.get("data", {})
                        
                        if msg_type == "progress":
                            progress = msg_data.get("value", 0) / max(msg_data.get("max", 1), 1)
                            self._tasks[prompt_id]["progress"] = progress
                            self._tasks[prompt_id]["status"] = "running"
                            yield {
                                "code": SERVER_CODE_OK,
                                "data": {
                                    "type": "progress",
                                    "progress": progress,
                                    "step": msg_data.get("value"),
                                    "total_steps": msg_data.get("max"),
                                    "node": msg_data.get("node")
                                }
                            }
                        
                        elif msg_type == "executing":
                            node = msg_data.get("node")
                            if node is None and msg_data.get("prompt_id") == prompt_id:
                                self._tasks[prompt_id]["status"] = "completed"
                                break
                            else:
                                yield {
                                    "code": SERVER_CODE_OK,
                                    "data": {"type": "executing", "node": node}
                                }
                        
                        elif msg_type == "executed":
                            if msg_data.get("prompt_id") == prompt_id:
                                output = msg_data.get("output", {})
                                if "images" in output:
                                    yield {
                                        "code": SERVER_CODE_OK,
                                        "data": {
                                            "type": "node_output",
                                            "node": msg_data.get("node"),
                                            "images_info": output["images"]
                                        }
                                    }
                        
                        elif msg_type == "execution_error":
                            error_msg = msg_data.get("exception_message", "Unknown error")
                            logger.error(f"Execution error: {error_msg}")
                            self._tasks[prompt_id]["status"] = "error"
                            yield {"code": SERVER_CODE_ERROR, "data": error_msg}
                            return
                    
                    elif msg.type in (aiohttp.WSMsgType.ERROR, aiohttp.WSMsgType.CLOSED):
                        yield {"code": SERVER_CODE_ERROR, "data": "WebSocket error"}
                        return
            
            # 获取生成的图像
            images = await self._get_images(prompt_id)
            self._tasks[prompt_id]["images"] = images
            
            yield {
                "code": SERVER_CODE_OK,
                "data": {
                    "type": "result",
                    "images": images,  # base64 编码的图像列表
                    "prompt_id": prompt_id
                }
            }
            yield {"code": SERVER_CODE_STOP, "data": ""}
            
        except asyncio.TimeoutError:
            logger.error("Generation timeout")
            yield {"code": SERVER_CODE_ERROR, "data": "timeout"}
        except Exception as e:
            logger.error(f"Error: {e}")
            yield {"code": SERVER_CODE_ERROR, "data": str(e)}
        finally:
            cost = int(time.time() * 1000) - start
            logger.info(f"Generation completed, cost: {cost}ms")

    async def _queue_prompt(self, workflow: Dict[str, Any], client_id: str) -> Optional[str]:
        """提交工作流到 ComfyUI"""
        try:
            async with self._http_session.post(
                f"{self._comfyui_url}/prompt",
                json={"prompt": workflow, "client_id": client_id}
            ) as resp:
                if resp.status == 200:
                    result = await resp.json()
                    return result.get("prompt_id")
                else:
                    error = await resp.text()
                    uvicorn_logger.error(f"Failed to queue prompt: {error}")
                    return None
        except Exception as e:
            uvicorn_logger.error(f"Error queuing prompt: {e}")
            return None

    async def _get_images(self, prompt_id: str) -> List[str]:
        """获取生成的图像（base64）"""
        images = []
        try:
            async with self._http_session.get(f"{self._comfyui_url}/history/{prompt_id}") as resp:
                if resp.status == 200:
                    history = await resp.json()
                    if prompt_id in history:
                        outputs = history[prompt_id].get("outputs", {})
                        for node_id, node_output in outputs.items():
                            if "images" in node_output:
                                for img_info in node_output["images"]:
                                    params = {
                                        "filename": img_info["filename"],
                                        "subfolder": img_info.get("subfolder", ""),
                                        "type": img_info.get("type", "output")
                                    }
                                    async with self._http_session.get(
                                        f"{self._comfyui_url}/view",
                                        params=params
                                    ) as img_resp:
                                        if img_resp.status == 200:
                                            img_data = await img_resp.read()
                                            images.append(base64.b64encode(img_data).decode("utf-8"))
        except Exception as e:
            uvicorn_logger.error(f"Error getting images: {e}")
        return images